(**
 @file    automataUtil.ml
 @brief   Utility of automata
 @author  shinyak
 @version $Rev: 270 $

 Shinya KAWANAKA (C) 2004 All Rights Reserved.
 $Id: automataUtil.ml 270 2005-05-08 10:25:02Z shinyak $
*)

open Util
open Syntax

(* ---------------------------------------------------------------------- *)

module Symbol : sig
  type t = 
    | SymNone
    | SymAnyElemAttr               (* match any element and attribute *)
    | SymAnyElem                   (* match any element *)
    | SymAnyAttr                   (* match any attribute *)
    | SymAnyStr                    (* match any string *)
    | SymStr    of string          (* match string *)
    | SymLbl    of string          (* match label *)
    | SymAttr   of string          (* match attribute *)
    | SymAttrLs of string          (* match { x | x < string } *)
    | SymAttrGr of string          (* match { x | string < x } *)
    | SymAttrBt of string * string (* match { x | lhs < x && x < rhs } *)

  val anyelemattr_symbol : t
  val anyelem_symbol : t
  val anystr_symbol  : t
  val anyattr_symbol : t
  val str_symbol     : string -> t
  val lbl_symbol     : string -> t
  val attr_symbol    : string -> t
  val attrls_symbol  : string -> t
  val attrgr_symbol  : string -> t
  val attrbt_symbol  : string -> string -> t

  val matches        : t -> Syntax.xml_tree -> bool
  val unify_symbol   : t -> t -> t
  val matches_symbol : t -> t -> bool

  val compare        : t -> t -> int
  val to_string      : t -> string
end = struct
  type t =
    | SymNone
    | SymAnyElemAttr
    | SymAnyElem
    | SymAnyAttr
    | SymAnyStr
    | SymStr    of string
    | SymLbl    of string
    | SymAttr   of string
    | SymAttrLs of string
    | SymAttrGr of string
    | SymAttrBt of string * string

  let anyelemattr_symbol = SymAnyElemAttr
  let anyelem_symbol = SymAnyElem
  let anystr_symbol = SymAnyStr
  let anyattr_symbol = SymAnyAttr
  let str_symbol str = SymStr str
  let lbl_symbol lbl = SymLbl lbl
  let attr_symbol attr = SymAttr attr
  let attrls_symbol attr = SymAttrLs attr
  let attrgr_symbol attr = SymAttrGr attr
  let attrbt_symbol ls gt = SymAttrBt (ls, gt)

  (** if [symbol] matches [xmltree], return true, otherwise false *)
  let matches symbol xmltree = match (symbol, xmltree) with
    | (SymNone, _) -> false

    | (SymAnyElemAttr, XMLLeaf) -> false
    | (SymAnyElemAttr, _)       -> true

    | (SymAnyElem, XMLLeaf)     -> false
    | (SymAnyElem, XMLString _) -> true
    | (SymAnyElem, XMLAttr _)   -> false
    | (SymAnyElem, XMLNode _)   -> true

    | (SymAnyAttr, XMLAttr _)   -> true
    | (SymAnyAttr, _)           -> false

    | (SymAnyStr, XMLString _)  -> true
    | (SymAnyStr, _)            -> false

    | (SymStr str, XMLString (s, _,  _)) -> str = s
    | (SymStr _,   _)                    -> false

    | (SymLbl lbl, XMLNode (s, _, _)) -> lbl = s
    | (SymLbl _, _)                   -> false

    | (SymAttr attr, XMLAttr (s, _, _)) -> attr = s
    | (SymAttr _, _)                    -> false

    | (SymAttrLs attr, XMLAttr (s, _, _)) -> s < attr
    | (SymAttrLs _, _)                    -> false

    | (SymAttrGr attr, XMLAttr (s, _, _)) -> attr < s
    | (SymAttrGr _, _)                    -> false

    | (SymAttrBt (ls, gt), XMLAttr (s, _, _)) -> ls < s && s < gt
    | (SymAttrBt _, _)                        -> false


  let unify_symbol sym1 sym2 = match (sym1, sym2) with
    | (SymNone, _) -> SymNone
    | (_, SymNone) -> SymNone

    | (SymAnyElemAttr, _) -> sym2
    | (_, SymAnyElemAttr) -> sym1

    | (SymAnyElem, SymAnyElem) -> SymAnyElem
    | (SymAnyElem, SymAnyStr)  -> SymAnyStr
    | (SymAnyElem, SymStr _)   -> sym2
    | (SymAnyElem, SymLbl _)   -> sym2
    | (SymAnyElem, _)          -> SymNone

    | (SymAnyAttr, SymAnyAttr)  -> SymAnyAttr
    | (SymAnyAttr, SymAttr _)   -> sym2
    | (SymAnyAttr, SymAttrLs _) -> sym2
    | (SymAnyAttr, SymAttrGr _) -> sym2
    | (SymAnyAttr, SymAttrBt _) -> sym2
    | (SymAnyAttr, _)           -> SymNone

    | (SymAnyStr, SymAnyElem) -> SymAnyStr
    | (SymAnyStr, SymAnyStr)  -> SymAnyStr
    | (SymAnyStr, SymStr _)   -> sym2
    | (SymAnyStr, _)          -> SymNone

    | (SymStr str, SymAnyElem)  -> sym1
    | (SymStr str, SymAnyStr)   -> sym2
    | (SymStr str, SymStr str2) -> if str = str2 then sym1 else SymNone
    | (SymStr str, _)           -> SymNone

    | (SymLbl lbl, SymAnyElem)  -> sym1 
    | (SymLbl lbl, SymLbl lbl2) -> if lbl = lbl2 then sym1 else SymNone
    | (SymLbl lbl, _)           -> SymNone

    | (SymAttr attr, SymAnyAttr)         -> sym1
    | (SymAttr attr, SymAttr attr2)      -> if attr = attr2 then sym1 else SymNone
    | (SymAttr attr, SymAttrLs ls)       -> if attr < ls then sym1 else SymNone
    | (SymAttr attr, SymAttrGr gr)       -> if gr < attr then sym1 else SymNone
    | (SymAttr attr, SymAttrBt (l, r))   -> if l < attr && attr < r then sym1 else SymNone
    | (SymAttr attr, _)                  -> SymNone


    | (SymAttrLs attr, SymAnyAttr)         -> sym1
    | (SymAttrLs attr, SymAttr attr2)      -> if attr2 < attr then sym2 else SymNone
    | (SymAttrLs attr, SymAttrLs ls)       -> SymAttrLs (if attr < ls then attr else ls)
    | (SymAttrLs attr, SymAttrGr gr)       -> if (attr > gr) then SymAttrBt (gr, attr) else SymNone
    | (SymAttrLs attr, SymAttrBt (l, r))   -> 
        if attr < l then      SymNone
        else if attr < r then SymAttrBt (l, attr)
        else                  sym2
    | (SymAttrLs attr, _)                  -> SymNone

    | (SymAttrGr attr, SymAnyAttr)         -> sym1
    | (SymAttrGr attr, SymAttr attr2)      -> if attr < attr2 then sym2 else SymNone
    | (SymAttrGr attr, SymAttrLs ls)       -> if attr < ls then SymAttrBt (attr, ls) else SymNone
    | (SymAttrGr attr, SymAttrGr gr)       -> SymAttrGr (if attr < gr then gr else attr)
    | (SymAttrGr attr, SymAttrBt (l, r))   -> 
        if r < attr then      SymNone
        else if l < attr then SymAttrBt (attr, r)
        else                  sym2
    | (SymAttrGr attr, _)                  -> SymNone


    | (SymAttrBt (ls, gr), SymAnyAttr)           -> sym1
    | (SymAttrBt (ls, gr), SymAttr attr)         -> if ls < attr && attr < gr then sym2 else SymNone
    | (SymAttrBt (ls, gr), SymAttrLs ls2)        ->
        if ls2 < ls then      SymNone
        else if ls2 < gr then SymAttrBt (ls, ls2)
        else                  sym1
    | (SymAttrBt (ls, gr), SymAttrGr gr2)        ->
        if gr < gr2 then      SymNone
        else if ls < gr2 then SymAttrBt (gr2, gr)
        else                  sym1
    | (SymAttrBt (ls, gr), SymAttrBt (ls2, gr2)) ->
        let lsmax = if ls < ls2 then ls2 else ls
        and grmin = if gr < gr2 then gr else gr2 in
        if lsmax > grmin then SymNone
        else SymAttrBt (lsmax, grmin)
    | (SymAttrBt (ls, gr), _)                    -> SymNone


  let matches_symbol sym1 sym2 =
    let xmltree_of_symbol = function 
      | SymNone -> XMLLeaf
      | SymAnyElemAttr
      | SymAnyElem
      | SymAnyAttr
      | SymAnyStr -> invalid_arg "match_symbol"
      | SymStr str -> XMLString (str, XMLLeaf, XMLLeaf)
      | SymLbl lbl -> XMLNode (lbl, XMLLeaf, XMLLeaf)
      | SymAttr attr -> XMLAttr (attr, XMLLeaf, XMLLeaf)
      | SymAttrLs _
      | SymAttrGr _
      | SymAttrBt _ -> invalid_arg "match_symbol"
    in
    matches sym1 (xmltree_of_symbol sym2)

  let compare = compare
  let to_string x = 
    match x with
      | SymNone              -> "SymNone"
      | SymAnyElemAttr       -> "SymAnyElemAttr"
      | SymAnyElem           -> "SymAnyElem"
      | SymAnyAttr           -> "SymAnyAttr"
      | SymAnyStr            -> "SymAnyStr"
      | SymStr str           -> "SymStr" ^ str     
      | SymLbl lbl           -> "SymLbl:" ^ lbl
      | SymAttr attr         -> "SymAttr:" ^ attr
      | SymAttrLs attr       -> "SymAttr:x<" ^ attr
      | SymAttrGr attr       -> "SymAttr:" ^ attr ^ "<x"
      | SymAttrBt (lhs, rhs) -> "SymAttr:" ^ lhs ^ "<x<" ^ rhs
end

module SymbolMap = Map.Make(Symbol)
module SymbolSet = Set.Make(Symbol)
type symbol = Symbol.t
type symbol_set = SymbolSet.t

(* ---------------------------------------------------------------------- *)

module rec Variable : sig
  type t = 
    | VarBindBegin     of ident
    | VarBindEnd       of ident
    | VarAutomataBegin of ident
    | VarAutomataEnd   of ident
    | VarKill          of ident

  val compare          : t -> t -> int
  val equal            : t -> t -> bool

  val create_bindbegin : string -> t
  val create_bindend   : string -> t
  val create_automatabegin : string -> t
  val create_automataend   : string -> t
  val create_kill      : ident -> t

  val to_string        : t -> string
  val to_symbol        : t -> string
end = struct
  type t = 
    | VarBindBegin     of ident
    | VarBindEnd       of ident
    | VarAutomataBegin of ident
    | VarAutomataEnd   of ident
    | VarKill          of ident

  let compare = compare
  let equal v1 v2 = v1 = v2

  let create_bindbegin ident     = VarBindBegin     ident
  let create_bindend ident       = VarBindEnd       ident
  let create_automatabegin ident = VarAutomataBegin ident
  let create_automataend   ident = VarAutomataEnd   ident
  let create_kill ident          = VarKill ident

  let to_string = function
    | VarBindBegin x     -> "B"  ^ x
    | VarBindEnd   x     -> "~B" ^ x
    | VarAutomataBegin x -> "A"  ^ x 
    | VarAutomataEnd   x -> "~A" ^ x
    | VarKill x          -> "K"  ^ x

  let to_symbol = function
    | VarBindBegin x
    | VarBindEnd   x 
    | VarAutomataBegin x 
    | VarAutomataEnd x 
    | VarKill x          -> x
end
and State : sig
  type t

  val compare                  : t -> t -> int
  val create                   : unit -> t

  val id                       : t -> int
  val equiv                    : t -> t -> bool
    
  val set_as_final             : t -> bool -> unit
  val is_final                 : t -> bool

  val push_front_transition    : t -> Transition.t -> unit
  val append_front_transitions : t -> Transition.t list -> unit
  val append_back_transitions  : t -> Transition.t list -> unit
  val change_transitions       : t -> Transition.t list -> unit
  val get_transitions          : t -> Transition.t list

  val push_front_epsilon       : t -> EpsilonTransition.t -> unit
  val append_front_epsilons    : t -> EpsilonTransition.t list -> unit
  val append_back_epsilons     : t -> EpsilonTransition.t list -> unit
  val change_epsilons          : t -> EpsilonTransition.t list -> unit
  val get_epsilons             : t -> EpsilonTransition.t list


  val push_front_variable      : t -> Variable.t -> unit
  val append_front_variables   : t -> Variable.t list -> unit
  val append_back_variables    : t -> Variable.t list -> unit
  val change_variables         : t -> Variable.t list -> unit
  val get_variables            : t -> Variable.t list

  (** create a new state by copying. *)
  val create_by_copy           : t -> t
  (** create shallow copy *)
  val shallow_copy             : t -> t

end = struct
  type t = {
    id                  : int;
    mutable final       : bool;
    mutable variables   : Variable.t list;
    mutable transitions : Transition.t list;
    mutable epsilons    : EpsilonTransition.t list;
  }

  let compare = compare

  let create =
    let id_counter = ref 0 in
    fun () ->
      incr id_counter;
      {
        id          = !id_counter;
        final       = false;
        variables   = [];
        transitions = [];
        epsilons    = [];
      }

  let id t = t.id

  let equiv t1 t2 = 
    try
      t1.final       = t2.final     &&
      List.for_all2 Variable.equal t1.variables t2.variables &&
      List.for_all2 Transition.equal t1.transitions t2.transitions
    with
      Invalid_argument _ -> false

  (** set if this state is final state or not. *)
  let set_as_final t b =
    t.final <- b

  (** get if this state is final state or not. *)
  let is_final t =
    t.final

  let push_front_transition t trans =
    t.transitions <- trans :: t.transitions

  let append_front_transitions t transs =
    t.transitions <- transs @ t.transitions

  let append_back_transitions t transs =
    t.transitions <- t.transitions @ transs

  let change_transitions t transs =
    t.transitions <- transs

  let get_transitions t =
    t.transitions

  let push_front_epsilon t eps =
    t.epsilons <- eps :: t.epsilons

  let append_front_epsilons t epss =
    t.epsilons <- epss @ t.epsilons

  let append_back_epsilons t epss =
    t.epsilons <- t.epsilons @ epss

  let change_epsilons t epss =
    t.epsilons <- epss

  let get_epsilons t =
    t.epsilons

  let push_front_variable t var =
    t.variables <- var :: t.variables

  let append_front_variables t vars =
    t.variables <- vars @ t.variables

  let append_back_variables t vars =
    t.variables <- t.variables @ vars

  let change_variables t vars =
    t.variables <- vars

  let get_variables t =
    t.variables
      
  let create_by_copy t =
    let s = create () in
    s.final <- t.final;
    s.variables <- t.variables;
    s.transitions <- t.transitions;
    s.epsilons <- t.epsilons;
    s

  let shallow_copy t = {
    id          = t.id;
    final       = t.final;
    variables   = t.variables;
    transitions = t.transitions;
    epsilons    = t.epsilons;
  }

end

and Transition : sig
  type t = symbol * State.t * State.t list
  val compare     : t -> t -> int
  val equal       : t -> t -> bool
  val create      : Symbol.t -> State.t -> State.t list -> Symbol.t * State.t * State.t list
  val symbol      : t -> symbol
  val child_state : t -> State.t
  val next_state  : t -> State.t list
end = struct
  type t                    = symbol * State.t * State.t list
  let compare               = compare
  let equal (sym1, sta1, stb1) (sym2, sta2, stb2) =
    sym1 = sym2 && (State.id sta1 = State.id sta2) && List.for_all2 (fun a b -> (State.id a = State.id b)) stb1 stb2
    
  let create symbol s1 s2   = (symbol, s1, s2)
  let symbol (x, _, _)      = x
  let child_state (_, x, _) = x
  let next_state (_, _, x)  = x
end

and EpsilonTransition : sig
  type t
  val compare : t -> t -> int
  val create  : State.t -> t
  val next    : t -> State.t
end = struct
  type t           = State.t
  let compare      = compare
  let create state = state
  let next t       = t
end
;;

(* ---------------------------------------------------------------------- *)

module VariableSet = Set.Make(Variable)
module VariableMap = Map.Make(Variable)
module StateSet = Set.Make(State)
module StateMap = Map.Make(State)
module TransitionSet = Set.Make(Transition)
module EpsilonTransitionSet = Set.Make(EpsilonTransition)
type variable = Variable.t
type variable_set = VariableSet.t
type state = State.t
type state_set = StateSet.t
type transition = Transition.t
type transition_set = TransitionSet.t
type epsilon_transition = EpsilonTransition.t
type epsilon_transition_set = EpsilonTransitionSet.t


let state_set_of_state_list lst =
  let rec iter set = function
      [] -> set
    | l :: ls -> iter (StateSet.add l set) ls
  in
  iter StateSet.empty lst
;;

