(**
 @file    binding.ml
 @brief   Implementation of binding
 @author  shinyak
 @version $Rev$

 Shinya KAWANAKA (C) 2004 All Rights Reserved.
 $Id$
*)

open Util
open AutomataUtil
open MarkingAutomata

module Result : sig
  type r =
    | BindingRevList of Syntax.xml_rev_tree list     (* binding *)
    | InterRevList   of t list                       (* automata relation *)
  and t = r StringMap.t                              (* mapping from variable *)

  val compare : t -> t -> int
  (* val prdbg : t -> unit *)
  val print : t -> unit
  val empty : t

end = struct

  type r = 
    | BindingRevList of Syntax.xml_rev_tree list
    | InterRevList   of t list
  and t = r StringMap.t

  let compare = compare

(*
  let rec prdbg result =
    StringMap.iter (fun t r ->
      prdbg_string t;
      prdbg_string " -> ";
      match r with
        | BindingRevList lst ->
            prdbg_string "[";
            List.iter (fun x -> ReadXML.prdbg_xmltree (Syntax.xml_tree_of_xml_rev_tree x)) lst;
            prdbg_string "]";
            prdbg_newline ()
        | InterRevList lst ->
            List.iter (fun x -> prdbg_endline "---->"; List.iter prdbg x) lst
    ) result
*)
  let rec print result =
    StringMap.iter (fun t r ->
      print_string t;
      print_string " -> ";
      match r with
        | BindingRevList lst ->
            print_string "[";
            List.iter (fun x -> ReadXML.print_xmltree (Syntax.xml_tree_of_xml_rev_tree x)) lst;
            print_string "]";
            print_newline ()
        | InterRevList lst ->
            List.iter (fun x -> print_endline "---->("; print x; print_endline ")") lst
    ) result

  let empty = StringMap.empty
end;;


module rec Configuration : sig

  type t = {
    state        : State.t;                      (**< current state *)
    bvar         : ident option;                 (**< variables for binding *)
    continuation : Continuation.t list;          (**< Continuation list *)
    parents      : (ident * Result.t) list;      (**< For another relation *)
    result       : Result.t;                     (**< set of result *)
  }

  val compare : t -> t -> int
  (* val prdbg : t -> unit *)

end = struct

  type t = {
    state        : State.t;
    bvar         : ident option;
    continuation : Continuation.t list;
    parents      : (ident * Result.t) list;
    result       : Result.t;
  }
  let compare = compare
(*
  let prdbg x =
    prdbg_endline  "------------------------------";
    (* prdbg_endline ("relname : " ^ x.relname); *)
    prdbg_endline ("state   : " ^ (string_of_int (State.id x.state)));
    prdbg_string  ("bvar    : "); (match x.bvar with Some y -> prdbg_string y; prdbg_char ' ' | None -> ()); prdbg_newline ();
    prdbg_string  ("cont    : "); (match x.continuation with | [] -> () | x :: _ -> 
      List.iter (fun y -> prdbg_int (State.id y); prdbg_char ' ') x.Continuation.state); prdbg_newline ();
    prdbg_string  ("parents : "); (match x.parents with | [] -> () | (x, _) :: _ -> prdbg_string x); prdbg_newline ();
(*(match x.parent with Some (_, y) -> prdbg_int (State.id y.state) | None -> ()); prdbg_newline ();*)
    prdbg_string  ("result  : "); List.iter (fun x -> Result.prdbg x; prdbg_char ' ') x.result; prdbg_newline ();
*)
end

and Continuation : sig
  type t = {
    state        : State.t list;
    bvar         : ident option;
  }
end = struct
  type t = {
    state        : State.t list;
    bvar         : ident option;
  }
end
;;


let create_binding (input_am : MarkingAutomata.t) (input_xml : Syntax.xml_tree) =
  
  (** create initial configurations from relation *)
  let create_initial_configurations () =
    let automata = input_am in
    List.map (fun initial_state ->
      (* prdbg_endline (string_of_int (State.id initial_state)); *)
      { Configuration.state        = initial_state;
        Configuration.bvar         = None;
        Configuration.continuation = [];
        Configuration.parents      = [];
        Configuration.result       = Result.empty;
      }
    ) (StateSet.elements automata.initials)
  in

  let rec create_child_configurations (xmltree : Syntax.xml_tree) (configurations : Configuration.t list) : Configuration.t list =
    List.flatten (List.map (fun cfg -> create_child_configuration xmltree cfg) configurations)
  and create_child_configuration (xmltree : Syntax.xml_tree) (cfg : Configuration.t) : Configuration.t list =
    (* prdbg_endline "[CCC]"; *)

    List.fold_right (fun transition child_lst ->
      let symbol = Transition.symbol transition in
      if Symbol.matches symbol xmltree then begin
        let continuation : Continuation.t =
          { Continuation.state    = Transition.next_state transition;
            Continuation.bvar     = cfg.Configuration.bvar;
          }
        in

        let child_val : Configuration.t = 
          let new_result = 
            let rslt = cfg.Configuration.result in
            match cfg.Configuration.bvar with
              | None -> rslt
              | Some elt -> begin
                  match StringMap.find elt rslt with
                    | Result.BindingRevList rev_tree_list ->
                        let rev_tree = List.hd rev_tree_list in
                        let new_rev_tree = match xmltree with
                          | Syntax.XMLNode (label, child, _) ->
                              Syntax.XMLRevNode (label, Syntax.xml_rev_tree_of_xml_tree child, rev_tree)
                          | Syntax.XMLAttr (attr, child, _) ->
                              Syntax.XMLRevAttr (attr, Syntax.xml_rev_tree_of_xml_tree child, rev_tree)
                          | Syntax.XMLString (str, child, _) ->
                              Syntax.XMLRevString (str, Syntax.xml_rev_tree_of_xml_tree child, rev_tree)
                          | _ ->
                              failwith "shouldn't happen"
                        in
                        let new_rev_tree_list = new_rev_tree :: (List.tl rev_tree_list) in
                        StringMap.add elt (Result.BindingRevList new_rev_tree_list) rslt
                    | Result.InterRevList _ -> 
                        failwith "something wrong"
                end
          in
          
          { Configuration.state        = Transition.child_state transition;
            Configuration.bvar         = None;
            Configuration.continuation = continuation :: cfg.Configuration.continuation;
            Configuration.parents      = cfg.Configuration.parents;
            Configuration.result       = new_result;
          }
        in

        child_val :: child_lst

      end else begin
        (* skip *)
        child_lst
      end
    ) (State.get_transitions cfg.Configuration.state) []
  in

  let rec pre_transition (configurations : Configuration.t list) : Configuration.t list =
    (* prdbg_endline "[PRE_TRANSITION]";
       List.iter (fun x -> prdbg_int (State.id x.Configuration.state); prdbg_char ' ') configurations;
       prdbg_newline (); *)
    let process_var (cfg : Configuration.t) (var : Variable.t) : Configuration.t list =
      match var with
        | Variable.VarBindBegin ident ->
            (* prdbg_endline ("[EVENT] " ^ (string_of_int (State.id cfg.Configuration.state)) ^ " VARBIND BEGIN " ^ ident); *)
            assert (cfg.Configuration.bvar = None);
            let new_result = 
              let result = cfg.Configuration.result in
              let newval = try 
                match StringMap.find ident result with
                  | Result.BindingRevList lst -> Result.BindingRevList (Syntax.XMLRevLeaf :: lst)
                  | Result.InterRevList _ -> failwith "shouldn't happen"
              with
                | Not_found -> Result.BindingRevList [Syntax.XMLRevLeaf]
              in
              StringMap.add ident newval result
            in
            [{ cfg with
              Configuration.bvar   = Some ident;
              Configuration.result = new_result;
            }]
        | Variable.VarBindEnd ident ->
            (* prdbg_endline ("[EVENT] " ^ (string_of_int (State.id cfg.Configuration.state)) ^ " VARBIND END " ^ ident); *)
            assert (cfg.Configuration.bvar = Some ident);
            [{ cfg with
              Configuration.bvar = None;
            }]
        | Variable.VarAutomataBegin ident ->
            (* prdbg_endline ("[EVENT] " ^ (string_of_int (State.id cfg.Configuration.state)) ^ " VARAUTOMATA BEGIN " ^ ident); *)
                           
            assert (cfg.Configuration.bvar = None);
            [{ cfg with
              Configuration.parents = (ident, cfg.Configuration.result) :: cfg.Configuration.parents;
              Configuration.result  = Result.empty
            }]
        | Variable.VarAutomataEnd ident ->
            (* prdbg_endline ("[EVENT] " ^ (string_of_int (State.id cfg.Configuration.state)) ^ " VARAUTOMATA END " ^ ident); *)
            begin
              match cfg.Configuration.parents with
                | [] -> failwith ("shouldn't happen: parent " ^ ident ^ " " ^ (string_of_int (State.id cfg.Configuration.state)));
                | (avar, result) :: parents ->
                    assert (ident = avar);
                    let new_result = 
                      let r = result in
                      let t = try StringMap.find avar r with Not_found -> Result.InterRevList [] in
                      match t with
                        | Result.BindingRevList _ -> failwith "something wrong"
                        | Result.InterRevList s -> StringMap.add avar (Result.InterRevList (cfg.Configuration.result :: s)) r
                    in
                    [{ cfg with
                      Configuration.parents = parents;
                      Configuration.result  = new_result;
                    }]
            end
        | Variable.VarKill ident ->
            [cfg]
    in

    (* process_var -> epsilon_transition *) 
    let rec iter (cfg : Configuration.t) : Configuration.t list =
      let vars = State.get_variables cfg.Configuration.state in
      let var_processed_confs : Configuration.t list = 
        List.fold_left (fun confs var ->
          List.flatten (List.map (fun conf -> process_var conf var) confs)
        ) [cfg] vars
      in
      var_processed_confs
    in
    List.flatten (List.map (fun conf -> iter conf) configurations)
  in

  let rec convert (configurations : Configuration.t list) (input_xml : Syntax.xml_tree) : Configuration.t list =
    (* prdbg_endline "================================================================================"; *)

    let (configs, rest_xml) = step configurations input_xml in
    if configs = [] then begin
      [] (* you don't have to process any more. *)
    end else begin
      match rest_xml with
        | None -> configs
        | Some xml -> convert configs xml
    end
  and step (configurations : Configuration.t list) (input_xml : Syntax.xml_tree) : (Configuration.t list * Syntax.xml_tree option) =
    (* prerr_endline ("[COMMAND] step -- rest " ^ (string_of_int (List.length configurations))); *)
    let configurations = pre_transition configurations in
    (* List.iter Configuration.prdbg configurations; *)
    assert (configurations <> []);

(*
    begin match input_xml with
      | Syntax.XMLLeaf ->
          prerr_endline "input_xml is XMLLeaf";
      | Syntax.XMLString (str, _, _) ->
          prerr_endline ("input_xml is XMLString: " ^ str);
      | Syntax.XMLAttr (attr, _, _) ->
          prerr_endline ("input_xml is XMLAttr: " ^ attr);
      | Syntax.XMLNode (lbl, _, _) ->
          prerr_endline ("input_xml is XMLNode:" ^ lbl);
    end;
*)
    match input_xml with
      | Syntax.XMLLeaf ->
          let cfgs =
            List.fold_right (fun cfg rslt ->
              if State.is_final cfg.Configuration.state then begin
                match cfg.Configuration.continuation with
                  | [] -> cfg :: rslt
                  | cont :: conts ->
                      let cfgs =
                        (List.map (fun state ->
                          { Configuration.state        = state;
                            Configuration.bvar         = cont.Continuation.bvar;
                            Configuration.continuation = conts;
                            Configuration.parents      = cfg.Configuration.parents;
                            Configuration.result       = cfg.Configuration.result;
                          }
                        ) cont.Continuation.state) 
                      in
                      cfgs @ rslt
              end else begin
                rslt
              end
            ) configurations []
          in
          (cfgs, None)
      | Syntax.XMLString (_, child, after)
      | Syntax.XMLAttr (_, child, after)
      | Syntax.XMLNode (_, child, after) ->
          let child_cfgs = create_child_configurations input_xml configurations in
          if child_cfgs <> [] then begin
            let child_itered = convert child_cfgs child in
            if child_itered <> [] then begin
              (child_itered, Some after)
            end else begin
              ([], Some after)
            end
          end else begin
            ([], Some after)
          end
  in

  (* ... *)
  let initial_configurations = create_initial_configurations () in  
  let configurations = convert initial_configurations input_xml in
  if configurations = [] then
    failwith "No configuration left."
  else
    List.map (fun x -> x.Configuration.result) configurations
;;
