(**
   @file    markingAutomata.ml
   @brief   Implementation of markingAutomata
   @author  shinyak
   @version $Rev: 270 $

   Shinya KAWANAKA (C) 2004 All Rights Reserved.
   $Id: markingAutomata.ml 270 2005-05-08 10:25:02Z shinyak $
*)

open Util
open Syntax
open AutomataUtil

module StringMap = Map.Make(String)

type t = {
  initials    : state_set;     (**< initial states *)
  nexts       : state_set;     (**< final states which is the same level of the initial states *)
};;

type marking_automata = t


type nonterminal = State.t

type ampattern = 
  | REmptyElem
  | RAnyElem                        (* PAnyElem matches attribute *)
  | RAnyString
  | RString    of string
  | RAnyAttr
  | RAttr      of ident * nonterminal   (* pattern can be PStringAny or PString or PVar *)
  | RAttrLs    of ident             (* { x | x < ident } *)
  | RAttrGr    of ident             (* { x | ident < x } *)
  | RAttrLsGr  of ident * ident     (* { x | left < x < right } *)
  | RLabel     of ident * nonterminal
  | RStar      of ampattern
  | RSeq       of ampattern * ampattern
  | ROr        of ampattern * ampattern
  | RVar       of ident * ampattern
  | RVarRel    of ident * ampattern
;;


let rec varset_of_ampattern nonpat = function
  | REmptyElem   
  | RAnyElem                        (* PAnyElem matches attribute *)
  | RAnyString
  | RString _
  | RAnyAttr        -> StringSet.empty
  | RAttr (_, non)  -> varset_of_nonterminal nonpat non
  | RAttrLs _
  | RAttrGr _
  | RAttrLsGr _     -> StringSet.empty
  | RLabel (_, non) -> varset_of_nonterminal nonpat non
  | RStar pat       -> varset_of_ampattern nonpat pat
  | RSeq (lhs, rhs) -> StringSet.union (varset_of_ampattern nonpat lhs) (varset_of_ampattern nonpat rhs)
  | ROr (lhs, rhs)  -> StringSet.union (varset_of_ampattern nonpat lhs) (varset_of_ampattern nonpat rhs)
  | RVar (x, _)     -> StringSet.singleton x
  | RVarRel (x, _)  -> StringSet.singleton x
and varset_of_nonterminal nonpat nonterminal =
  varset_of_ampattern nonpat (List.assq nonterminal nonpat)
;;



let prerr_marking_automata (automata : t) : unit =
  prerr_endline "initials ---";
  StateSet.iter (fun x -> prerr_int (State.id x); prerr_char ' ') automata.initials;
  prerr_newline ();
  prerr_endline "nexts ---";
  StateSet.iter (fun x -> prerr_int (State.id x); prerr_char ' ') automata.nexts;
  prerr_newline ();

  let rec prerr_states processed = function 
      [] -> ()
    | s :: ss ->
        if StateSet.mem s processed then begin
          prerr_states processed ss
        end else begin
          let processed = StateSet.add s processed in
          prerr_string ("trans : " ^ (string_of_int (State.id s)) ^ " --> ");
          if State.is_final s then prerr_string "final ";
          List.iter (fun x -> prerr_string (Variable.to_string x); prerr_char ' ') (State.get_variables s);
          prerr_newline ();
          let unprocessed = List.fold_left (fun unp (label, child, nexts) ->
            prerr_string "           -> (";
            prerr_string (Symbol.to_string label);
            prerr_string ", ";
            prerr_int    (State.id child);
            prerr_string ", ";
            List.iter (fun x -> prerr_int (State.id x); prerr_char ' ') nexts;
            prerr_endline ")";
            let unp1 = if StateSet.mem child processed then unp  else child :: unp  in
            let unp2 = (List.filter (fun x -> not (StateSet.mem x processed)) nexts) @ unp1 in
            unp2
          ) ss (State.get_transitions s) in
          let unprocessed = List.fold_left (fun unp next ->
            prerr_string "          ->e (";
            prerr_int    (State.id (EpsilonTransition.next next));
            prerr_string ")";
            prerr_newline ();
            if StateSet.mem (EpsilonTransition.next next) processed then unp else (EpsilonTransition.next next) :: unp 
          ) unprocessed (State.get_epsilons s) in
          prerr_states processed unprocessed
        end
  in
  prerr_states StateSet.empty (StateSet.elements automata.initials)
;;

(* ---------------------------------------------------------------------- *)
let validate_automata t =
  ()
;;


(* ---------------------------------------------------------------------- *)
(* optimize automata. remove epsilon transiton. *)
let optimize am = 
  let mapping = 
    let rec state_list_of_state state =
      (* prdbg_endline ("state_list_of_state " ^ (string_of_int (State.id state))); *)
      let epsilons = State.get_epsilons state in
      let afters =
        List.flatten (List.map (fun eps ->
          state_list_of_state (EpsilonTransition.next eps)
        ) epsilons)
      in
      [state] :: (List.map (fun x -> state :: x) afters)
    in
    let new_state_of_state_list states : State.t option = 
      let last = List.hd (List.rev states) in
      if State.get_transitions last <> [] || State.is_final last then begin
        let variables = 
          List.fold_right (fun state vars ->
            (State.get_variables state) @ vars
          ) states []
        in
        let s = State.create () in
        State.change_variables s variables;
        State.change_transitions s (State.get_transitions last);
        State.set_as_final s (State.is_final last);
        Some s
      end else begin
        None
      end
    in
    let rec loop mapping = function
      | [] -> mapping 
      | s :: ss ->
          if StateMap.mem s mapping then begin
            loop mapping ss
          end else begin
            let lst = state_list_of_state s in
            let new_states : State.t list = List.fold_right (fun sl rslt ->
              match new_state_of_state_list sl with
                | None -> rslt
                | Some x -> x :: rslt
            ) lst [] in
            let ss = List.fold_right (fun eps rslt ->
              (EpsilonTransition.next eps) :: rslt
            ) (State.get_epsilons s) ss in
            let ss = List.fold_right (fun trans rslt ->
              (Transition.child_state trans) :: (Transition.next_state trans) @ rslt
            ) (State.get_transitions s) ss in
            loop (StateMap.add s new_states mapping) ss
          end
    in
    loop StateMap.empty (StateSet.elements am.initials) 
  in

(*
  StateMap.iter (fun k v ->
    prdbg_int (State.id k);
    prdbg_string " ->";
    List.iter (fun s ->
      prdbg_string " ";
      prdbg_int (State.id s);
    ) v;
    prdbg_newline ()
  ) mapping;
*)

  let automata = 
    let initials =
      let initials = 
        List.flatten (List.map (fun x ->
          StateMap.find x mapping
        ) (StateSet.elements am.initials))
      in
      List.fold_right (fun x set ->
        StateSet.add x set
      ) initials StateSet.empty 
    in
    let nexts = 
      let nexts = 
        List.flatten (List.map (fun x ->
          StateMap.find x mapping 
        ) (StateSet.elements am.nexts))
      in
      List.fold_right (fun x set ->
        StateSet.add x set
      ) nexts StateSet.empty
    in
    let states =
      StateMap.fold (fun k v r ->
        v @ r
      ) mapping []
    in
    (* to change transitions *)
    List.iter (fun s ->
      let transitions = 
        List.fold_right (fun trans rslt ->
          let child  = Transition.child_state trans
          and next   = Transition.next_state  trans
          and symbol = Transition.symbol      trans  in 
          let children  = StateMap.find child mapping
          and nexts = List.flatten (List.map (fun x -> StateMap.find x mapping) next) in
          let transitions = List.map (fun chd -> Transition.create symbol chd nexts) children in
          transitions @ rslt
        ) (State.get_transitions s) []
      in
      State.change_transitions s transitions;
    ) states;

    { initials   = initials;
      nexts      = nexts;
    }
  in
  automata
;;

(* ---------------------------------------------------------------------- *)

(**
 * 
 *)
let create_marking_automata_accepts_empty () =
  let state = State.create () in
  State.set_as_final state true;
  { initials   = StateSet.singleton state;
    nexts      = StateSet.singleton state;
  }
;;


let create_marking_automata_accepts_anyelem () =
  let s1 = State.create ()
  and s2 = State.create ()
  and s3 = State.create () in
  State.set_as_final s2 true;
  State.set_as_final s3 true;
  State.push_front_transition s1 (Transition.create Symbol.anyelem_symbol s2 [s3]);
  State.push_front_transition s2 (Transition.create Symbol.anyelemattr_symbol s2 [s2]);
  { initials   = StateSet.singleton s1;
    nexts      = StateSet.singleton s3;
  }
;;


let create_marking_automata_accepts symbol =
  let s1, s2, s3 = State.create (), State.create (), State.create () in
  State.set_as_final s2 true;
  State.set_as_final s3 true;
  State.push_front_transition s1 (Transition.create symbol s2 [s3]);
  { initials   = StateSet.singleton s1;
    nexts      = StateSet.singleton s3;
  }
;;

let create_marking_automata_accepts_with_str symbol str_symbol =
  let s1, s2, s3, s4, s5 = State.create (), State.create (), State.create (), State.create (), State.create () in
  State.set_as_final s3 true;
  State.set_as_final s4 true;
  State.set_as_final s5 true;
  State.push_front_transition s1 (Transition.create symbol s2 [s3]);
  State.push_front_transition s2 (Transition.create str_symbol s4 [s5]);
  { initials   = StateSet.singleton s1;
    nexts      = StateSet.singleton s3;
  }


let create_marking_automata_top nonpat ampattern_top =
  let left = ref nonpat in
  
  let rec create_marking_automata dontkill ampattern =
    let varset = varset_of_ampattern nonpat ampattern in
    let (killed, am) = create_marking_automata_sub dontkill ampattern in
    let to_kill = StringSet.diff varset (StringSet.union killed dontkill) in
    let vars = List.map (fun x -> Variable.create_kill x) (StringSet.elements to_kill) in
    StateSet.iter (fun next ->
      State.append_back_variables next vars;
    ) am.nexts;
    (StringSet.union killed to_kill, am)
  and create_marking_automata_sub dontkill ampattern =
    match ampattern with
      | REmptyElem ->
          (* prdbg_endline "empty elem"; *)
          (StringSet.empty, create_marking_automata_accepts_empty ())
            
      | RAnyElem ->
          (* prdbg_endline "any elem"; *)
          (StringSet.empty, create_marking_automata_accepts_anyelem ())
            
      | RAnyString ->
          (* prdbg_endline "anystring"; *)
          (StringSet.empty, create_marking_automata_accepts Symbol.anystr_symbol)
            
      | RString str ->
          (* prdbg_endline "PString"; *)
          (StringSet.empty, create_marking_automata_accepts (Symbol.str_symbol str))
            
      | RAnyAttr ->
          (* prdbg_endline "PAnyAttr"; *)
          (StringSet.empty, create_marking_automata_accepts_with_str Symbol.anyattr_symbol (Symbol.anystr_symbol))
            
      | RAttr (ident, nonterminal) ->
          (* prdbg_endline "PAttr"; *)
          begin
            try 
              let pat = List.assq nonterminal !left in
              left := List.remove_assq nonterminal !left;
              let (killed, am) = create_marking_automata dontkill pat in
              StateSet.iter (fun s ->
                State.push_front_epsilon nonterminal (EpsilonTransition.create s)
              ) am.initials;
            with
              | Not_found -> ()
          end;

          let killed = StringSet.diff (varset_of_nonterminal nonpat nonterminal) dontkill in
          let child_vars = varset_of_nonterminal nonpat nonterminal in
          
          let s1, s2 = State.create (), State.create () in
          let transitions = [Transition.create (Symbol.attr_symbol ident) nonterminal [s2]] in
          State.change_transitions s1 transitions;
          State.set_as_final s2 true;
          
          let result_am = {
            initials   = StateSet.singleton s1;
            nexts      = StateSet.singleton s2;
          } in
          (killed, result_am)
            
      | RAttrLs ident ->
          (* prdbg_endline "PAttrLs"; *)
          (StringSet.empty, create_marking_automata_accepts_with_str (Symbol.attrls_symbol ident) (Symbol.anystr_symbol))
            
      | RAttrGr ident ->
          (* prdbg_endline "PAttrGr"; *)
          (StringSet.empty, create_marking_automata_accepts_with_str (Symbol.attrgr_symbol ident) (Symbol.anystr_symbol))
            
      | RAttrLsGr (ident_ls, ident_gr) ->
          (* prdbg_endline "PAttrLsGr"; *)
          (StringSet.empty, create_marking_automata_accepts_with_str (Symbol.attrbt_symbol ident_ls ident_gr) (Symbol.anystr_symbol))
            
      | RLabel (ident, nonterminal) ->
          
          (* prdbg_endline "PAttr"; *)
          begin
            try 
              let pat = List.assq nonterminal !left in
              left := List.remove_assq nonterminal !left;
              let (killed, am) = create_marking_automata dontkill pat in
              StateSet.iter (fun s ->
                State.push_front_epsilon nonterminal (EpsilonTransition.create s)
              ) am.initials;
            with
              | Not_found -> ()
          end;

          let killed = StringSet.diff (varset_of_nonterminal nonpat nonterminal) dontkill in
          let child_vars = varset_of_nonterminal nonpat nonterminal in
          
          let s1, s2 = State.create (), State.create () in
          let transitions = [Transition.create (Symbol.lbl_symbol ident) nonterminal [s2]] in
          State.change_transitions s1 transitions;
          State.set_as_final s2 true;
          
          let result_am = {
            initials   = StateSet.singleton s1;
            nexts      = StateSet.singleton s2;
          } in
          (killed, result_am)
            
      | RStar pat ->
          (* prdbg_endline "star"; *)
          let child_vars = varset_of_ampattern nonpat pat in
          let (killed, child_am) = create_marking_automata (StringSet.union child_vars dontkill) pat in

          assert (StringSet.is_empty killed);
          
          let new_initial_state = State.create () in
          State.set_as_final new_initial_state true;
          
          let s, s' = State.create (), State.create () in
          State.push_front_epsilon s (EpsilonTransition.create s');
          State.set_as_final s' true;
          StateSet.iter (fun init ->
            State.push_front_epsilon s (EpsilonTransition.create init)
          ) child_am.initials;
          StateSet.iter (fun next ->
            State.push_front_epsilon next (EpsilonTransition.create s');
            State.set_as_final next false
          ) child_am.nexts;
          StateSet.iter (fun init ->
            StateSet.iter (fun next ->
              State.push_front_epsilon next (EpsilonTransition.create init)
            ) child_am.nexts;
          ) child_am.initials;
          
          let result_am = {
            initials = StateSet.singleton s;
            nexts    = StateSet.singleton s';
          } in
          (killed, result_am)
            
      | RSeq (lhs, rhs) ->
          
          (* prdbg_endline "seq"; *)
          let lhs_vars = varset_of_ampattern nonpat lhs
          and rhs_vars = varset_of_ampattern nonpat rhs in
          
          let (lhs_killed, lhs_am) = create_marking_automata (StringSet.union rhs_vars dontkill) lhs
          and (rhs_killed, rhs_am) = create_marking_automata dontkill rhs in
          let killed = StringSet.union lhs_killed rhs_killed in
          
          let s = State.create () in
          StateSet.iter (fun next ->
            State.set_as_final next false;
            State.push_front_epsilon next (EpsilonTransition.create s);
          ) lhs_am.nexts;
          StateSet.iter (fun init ->
            State.push_front_epsilon s (EpsilonTransition.create init);
          ) rhs_am.initials;
          
          let result_am = {
            initials = lhs_am.initials;
            nexts    = rhs_am.nexts;
          } in
          
          (killed, result_am)
            
      | ROr (lhs, rhs) ->
          let lhs_vars = varset_of_ampattern nonpat lhs
          and rhs_vars = varset_of_ampattern nonpat rhs in
          
          let (lhs_killed, lhs_am) = create_marking_automata dontkill lhs
          and (rhs_killed, rhs_am) = create_marking_automata dontkill rhs in
          let killed = StringSet.inter lhs_killed rhs_killed in (* note: not union but inter *)
                    
          let result_am = {
            initials   = StateSet.union lhs_am.initials rhs_am.initials;
            nexts      = StateSet.union lhs_am.nexts rhs_am.nexts;
          } in
          (killed, result_am)            
            
      | RVar (ident, pat) ->
          (* This variable is for binding *)
          
          (* prdbg_endline "var"; *)
          assert (StringSet.is_empty (varset_of_ampattern nonpat pat));
          
          let (killed, am) = create_marking_automata dontkill pat in
          assert (StringSet.is_empty killed);
          
          StateSet.iter (fun next ->
            State.push_front_variable next (Variable.create_bindend ident);
          ) am.nexts;
                    
          StateSet.iter (fun initial ->
            State.push_front_variable initial (Variable.create_bindbegin ident)
          ) am.initials;
          
          (StringSet.empty, am)
      | RVarRel (ident, pat) ->
          (* This variable is for relation *)
          let (_, am) = create_marking_automata StringSet.empty pat in
          let ini, next = State.create (), State.create () in
          StateSet.iter (fun s ->
            State.push_front_epsilon ini (EpsilonTransition.create s)
          ) am.initials;
          StateSet.iter (fun s ->
            State.set_as_final s false;
            State.push_front_epsilon s (EpsilonTransition.create next)
          ) am.nexts;
          State.set_as_final next true;
          State.push_front_variable ini (Variable.create_automatabegin ident);
          State.push_front_variable next (Variable.create_automataend ident);

          let result_am = {
            initials   = StateSet.singleton ini;
            nexts      = StateSet.singleton next;
          } in
          (StringSet.empty, result_am)
  in

  snd (create_marking_automata StringSet.empty ampattern_top)
;;


(* return a list of nontermial * ampattern *)
let am_of_relations is_left toprelname relations =
  let top_nonterminal = State.create () in
  let top_relation = 
    try
      List.find (fun rel -> rel.name = toprelname) relations
    with
      | Not_found -> failwith ("toprelname (" ^ toprelname ^ ") is not found.")
  in
  let top_pattern = if is_left then top_relation.lhs else top_relation.rhs in

  let to_process : (nonterminal * relation * pattern) list ref   = ref [(top_nonterminal, top_relation, top_pattern)] in
  let processed  : (nonterminal * ampattern) list ref = ref [] in
  let idnon      : (pattern * nonterminal) list ref   = ref [] in

  let rec conv rel = function
    | PEmptyElem        -> REmptyElem
    | PAnyElem          -> RAnyElem
    | PAnyString        -> RAnyString
    | PString str       -> RString str
    | PAnyAttr          -> RAnyAttr 
    | PAttr (att, pat)  ->
        begin
          try 
            let nonterminal = List.assq pat !idnon in
            RAttr (att, nonterminal)
          with
            | Not_found ->
                let nonterminal = State.create () in
                to_process := (nonterminal, rel, pat) :: !to_process;
                idnon      := (pat, nonterminal) :: !idnon;
                RAttr (att, nonterminal)
        end
    | PAttrLs x         -> RAttrLs x
    | PAttrGr x         -> RAttrGr x
    | PAttrLsGr (x, y)  -> RAttrLsGr (x, y)
    | PLabel (lbl, pat) ->
        begin
          try 
            let nonterminal = List.assq pat !idnon in
            RLabel (lbl, nonterminal)
          with
            | Not_found ->
                let nonterminal = State.create () in
                to_process := (nonterminal, rel, pat) :: !to_process;
                idnon      := (pat, nonterminal) :: !idnon;
                RLabel (lbl, nonterminal)
        end
    | PStar pat         -> RStar (conv rel pat)
    | PSeq (lhs, rhs)   -> RSeq (conv rel lhs, conv rel rhs)
    | POr (lhs, rhs)    -> ROr (conv rel lhs, conv rel rhs)
    | PVar (x, pat)     -> RVar (x, conv rel pat)
    | PVarAny x         ->
        let (newrelname, _, _) = 
          try 
            List.find (fun (r, lhs, rhs) -> if is_left then x = lhs else x = rhs) rel.where 
          with
            | Not_found -> failwith (x ^ " is not found.")
        in
        let newrel = List.find (fun r -> r.name = newrelname) relations in
        let newpat = if is_left then newrel.lhs else newrel.rhs in
        RVarRel (x, conv newrel newpat)
  in

  while !to_process <> [] do
    let (nonterminal, relation, pattern) = List.hd !to_process in
    to_process := List.tl !to_process;
    let converted = conv relation pattern in
    processed := (nonterminal, converted) :: !processed;
  done;

  let top_ampat = List.assq top_nonterminal !processed in
  let am = create_marking_automata_top !processed top_ampat in
  StateSet.iter (fun ini ->
    State.push_front_epsilon top_nonterminal (EpsilonTransition.create ini)
  ) am.initials;
  { initials = StateSet.singleton top_nonterminal;
    nexts    = am.nexts;
  }
;;



let marking_automata_of_relations relname relations =
  let lhs_am = am_of_relations true relname relations
  and rhs_am = am_of_relations false relname relations in
  validate_automata lhs_am;
  validate_automata rhs_am;
  (optimize lhs_am, optimize rhs_am)
;;

