(**
 @file    output.ml
 @brief   Implementation of automata
 @author  shinyak
 @version $Rev$

 Shinya KAWANAKA (C) 2004 All Rights Reserved.
 $Id$
*)

open Util
open Syntax
open AutomataUtil
open MarkingAutomata
open Convert

module rec Configuration : sig
  type t = {
    state        : State.t;
    verif        : (ident * xml_tree) option;
    parents      : (ident * InputBinding.t) list;     (**< avar * parent binding *)
    binding      : InputBinding.t;
    continuation : Continuation.t list;
    result       : xml_rev_tree;
  }
  val compare : t -> t -> int
  (* val prdbg : t -> unit *)
end = struct 
  type t = {
    state        : State.t;
    verif        : (ident * xml_tree) option;
    parents      : (ident * InputBinding.t) list;
    binding      : InputBinding.t;
    continuation : Continuation.t list;
    result       : xml_rev_tree;
  }
  let compare = compare
  (*  
  let prdbg x =
    prdbg_endline  "------------------------------";
    prdbg_endline ("state        : " ^ (string_of_int (State.id x.state)));
    (* prdbg_string   "bvar         : "; (match x.bvar with None -> prdbg_newline () | Some y -> prdbg_endline y);*)
    prdbg_string   "verif        : "; (match x.verif with None -> prdbg_newline () | Some (y, tree) -> prdbg_string y; 
      prdbg_string " "; match tree with XMLLeaf -> prdbg_endline "leaf" | _ -> prdbg_endline "not_leaf");
    prdbg_string   "parents      : "; (match x.parents with [] -> prdbg_newline () | (y, z) :: _ -> prdbg_endline y);
    prdbg_string   "binding      : "; (if StringMap.is_empty x.binding then prdbg_endline "empty" else prdbg_endline "not empty");  (*StringMap.iter (fun k (InputBinding.InterList v) -> prdbg_string k; prdbg_string "->";  prdbg_int (List.length v)) x.binding; prdbg_newline (); *)
    prdbg_string   "continuation : "; (match x.continuation with [] -> prdbg_newline () | y :: _ -> List.iter (fun x ->
      prdbg_int (State.id x); prdbg_char ' ') y.Continuation.states; prdbg_newline ());
    prdbg_endline  "result       : (ry";
  *)  
end
and Continuation : sig
  type t = {
    states       : State.t list;
    verif        : (ident * xml_tree) option;
    result       : xml_rev_tree;
  }
end = struct
  type t = {
    states       : State.t list;
    verif        : (ident * xml_tree) option;
    result       : xml_rev_tree;
  }
end;;

let output_xml (output_am : MarkingAutomata.t) (input_binding : InputBinding.t list) : xml_tree =

  (* ---------------------------------------------------------------------- *)
  let create_initial_configurations (bindings : InputBinding.t list) : Configuration.t list =
    let automata = output_am in
    List.flatten (List.map (fun binding ->
      List.map (fun initial -> 
        { Configuration.state        = initial;
          Configuration.verif        = None;
          Configuration.parents      = [];
          Configuration.binding      = binding;
          Configuration.continuation = [];
          Configuration.result       = XMLRevLeaf;
        }
      ) (StateSet.elements automata.initials)
    ) bindings)
  in

  (* ---------------------------------------------------------------------- *)
  let rec create_child_configuration (cfg : Configuration.t) : Configuration.t list =
    (* prdbg_endline ("[CCC]: " ^ (Symbol.to_string lbl)); *)
    List.fold_right (fun (transition : Transition.t) (lst : Configuration.t list) ->
      let lbl = 
        match cfg.Configuration.verif with
          | None ->
              begin
                match Transition.symbol transition with
                  | Symbol.SymStr str -> Symbol.SymStr str
                  | Symbol.SymAttr attr -> Symbol.SymAttr attr 
                  | Symbol.SymLbl lbl -> Symbol.SymLbl lbl
                  | _ -> (* skip *) Symbol.SymNone
              end
          | Some (ident, xmltree) ->
              begin
                match xmltree with
                  | XMLLeaf -> Symbol.SymNone
                  | XMLNode (lbl, _, _) -> Symbol.SymLbl lbl
                  | XMLAttr (lbl, _, _) -> Symbol.SymAttr lbl
                  | XMLString (lbl, _, _) -> Symbol.SymStr lbl
              end
      in
      let symbol = Transition.symbol transition in
      if Symbol.matches_symbol symbol lbl then begin
        let (verif_child, verif_after) =
          match cfg.Configuration.verif with
            | None -> (None, None)
            | Some (ident, xml) ->
                match xml with
                  | XMLLeaf -> failwith "[CCC]: shouldn't happen"
                  | XMLNode (_, child, after)
                  | XMLAttr (_, child, after)
                  | XMLString (_, child, after) -> (Some (ident, child), Some (ident, after))
        in
        let cont_result = 
          match lbl with
            | Symbol.SymLbl lbl -> 
                XMLRevNode (lbl, XMLRevLeaf, cfg.Configuration.result)
            | Symbol.SymAttr attr ->
                XMLRevAttr (attr, XMLRevLeaf, cfg.Configuration.result)
            | Symbol.SymStr str ->
                XMLRevString (str, XMLRevLeaf, cfg.Configuration.result)
            | _ -> failwith "[CCC]: shouldn't happen 1"
        in
        let cont : Continuation.t =
          { Continuation.states       = Transition.next_state transition;
            Continuation.verif        = verif_after;
            Continuation.result       = cont_result;
          }
        in
        let child_val : Configuration.t =
          { Configuration.state        = Transition.child_state transition;
            Configuration.verif        = verif_child;
            Configuration.parents      = cfg.Configuration.parents;
            Configuration.binding      = cfg.Configuration.binding;
            Configuration.continuation = cont :: cfg.Configuration.continuation;
            Configuration.result       = XMLRevLeaf;
          }
        in
        child_val :: lst
      end else begin
        (* prdbg_endline "create_child_configuration: skipped"; *)
        lst
      end

    ) (State.get_transitions cfg.Configuration.state) []
  in

  (* ---------------------------------------------------------------------- *)
  let rec pre_transition (configurations : Configuration.t list) : (Configuration.t list * xml_tree list) =
    (* prdbg_endline "[PRE_TRANSITION]"; *)
    let process_var (cfg : Configuration.t) = function
      | Variable.VarBindBegin ident ->
          (* prdbg_endline ("[EVENT] VARBINDBEGIN " ^ ident); *)
          assert (cfg.Configuration.verif = None);
          begin try
            match (StringMap.find ident cfg.Configuration.binding) with
              | InputBinding.BindingList (v :: vs) ->
                  let new_binding = 
                    if vs = [] then StringMap.remove ident cfg.Configuration.binding
                    else StringMap.add ident (InputBinding.BindingList vs) cfg.Configuration.binding
                  in
                  [{ cfg with
                    Configuration.verif        = Some (ident, v);
                    Configuration.binding      = new_binding;
                  }]
              | _ -> failwith "shouldn't happen"
          with
            | Not_found -> []
          end
      | Variable.VarBindEnd ident ->
          (* prdbg_endline ("[EVENT] VARBINDEND " ^ ident); *)
          begin
            match cfg.Configuration.verif with
              | None -> failwith "shouldn't happen"
              | Some (_, XMLLeaf) ->
                  [{ cfg with
                    Configuration.verif        = None;
                  }]
              | _ -> []
          end
      | Variable.VarAutomataBegin ident ->
          (* prdbg_endline ("[EVENT] VARAUTOMATABEGIN " ^ ident); *)
          assert (cfg.Configuration.verif = None);
          begin try
            let ((binding, parent_binding) : (InputBinding.t * InputBinding.t)) = 
              match StringMap.find ident cfg.Configuration.binding with
                | InputBinding.InterList (bindings :: rest) ->
                    let parent_binding = 
                      if rest = [] then StringMap.remove ident cfg.Configuration.binding
                      else StringMap.add ident (InputBinding.InterList rest) cfg.Configuration.binding
                    in
                    (bindings, parent_binding)
                | _ -> failwith "shouldn't happen : VAB"
            in
            [{ cfg with
              Configuration.parents = (ident, parent_binding) :: cfg.Configuration.parents;
              Configuration.binding = binding;
            }]
          with
            | Not_found -> []
          end
      | Variable.VarAutomataEnd ident ->
          (* prdbg_endline ("[EVENT] VARAUTOMATAEND " ^ ident); *)
          assert (cfg.Configuration.verif = None);
          begin
            match cfg.Configuration.parents with
              | [] -> failwith "shouldn't happen" 
              | (avar, binding) :: parents ->
                  assert (avar = ident);
                  if StringMap.is_empty cfg.Configuration.binding then begin
                    [{ cfg with
                      Configuration.parents = parents;
                      Configuration.binding = binding;
                    }]
                  end else begin
                    (* prdbg_endline "VAE cut!";
                    StringMap.iter (fun k v -> prdbg_string k; prdbg_string " ") cfg.Configuration.binding;
                    prdbg_newline (); *)
                    []
                  end
          end
      | Variable.VarKill ident ->
          if StringMap.mem ident cfg.Configuration.binding then []
          else [cfg]
    in

    (* 1. process vars, 2. finalize, 3. epsilon transition *)
    let rec iter cfg : Configuration.t list * xml_tree list = 
      let var_processed : Configuration.t list =
        let vars = State.get_variables cfg.Configuration.state in
        List.fold_left (fun confs var ->
          List.flatten (List.map (fun conf -> process_var conf var) confs)
        ) [cfg] vars
      in
      (* prdbg_endline ("process_vars: " ^ (string_of_int (List.length var_processed))); *)
      let (finalized_cfgs, finalized_rslts) =
        List.fold_right (fun conf (finalized_cfgs, finalized_rslts) ->
          if State.is_final conf.Configuration.state (* && conf.Configuration.verif = None *) then begin
            let verif_ok = match conf.Configuration.verif with None -> true | Some (_, XMLLeaf) -> true | _ -> false in
            match conf.Configuration.continuation with
              | [] ->
                  if StringMap.is_empty conf.Configuration.binding && verif_ok then begin
                    (* prdbg_endline "OK!!!!"; *)
                    (finalized_cfgs, (xml_tree_of_xml_rev_tree conf.Configuration.result) :: finalized_rslts)
                  end else begin
                    (finalized_cfgs, finalized_rslts)
                  end
              | cont :: conts ->
                  if verif_ok then begin
                    let (* cfgs *) cfgrslts  = List.map (fun state -> 
                      let new_result = 
                        let revtree = conf.Configuration.result in
                        match cont.Continuation.result with
                          | XMLRevString (str, XMLRevLeaf, prev) ->
                              XMLRevString (str, revtree, prev)
                          | XMLRevAttr (attr, XMLRevLeaf, prev) ->
                              XMLRevAttr (attr, revtree, prev)
                          | XMLRevNode (lbl, XMLRevLeaf, prev) ->
                              XMLRevNode (lbl, revtree, prev)
                          | _ -> 
                              failwith "shouldn't happen"
                      in
                      iter {
                        Configuration.state        = state;
                        Configuration.verif        = cont.Continuation.verif;
                        Configuration.parents      = conf.Configuration.parents;
                        Configuration.binding      = conf.Configuration.binding;
                        Configuration.continuation = conts;
                        Configuration.result       = new_result;
                      }
                    ) cont.Continuation.states in
                    
                    List.fold_right (fun (cfgs, rslts) (cfgs_rslt, rslts_rslt) ->
                      ((cfgs @ cfgs_rslt), (rslts @ rslts_rslt))
                    ) cfgrslts (finalized_cfgs, finalized_rslts)
                    
                    (* ((cfgs @ finalized_cfgs), finalized_rslts) *)
                  end else begin
                    (finalized_cfgs, finalized_rslts)
                  end
          end else begin
            (finalized_cfgs, finalized_rslts)
          end
        ) var_processed ([], [])
      in
(*
      let (epsilon_transitted, epsilon_rslts) = 
        List.fold_right (fun conf (cfgs_rslt, rslt_rslt) ->
          let epsilons = State.get_epsilons conf.Configuration.state in
          if epsilons = [] then begin
            (conf :: cfgs_rslt, rslt_rslt)
          end else begin
            let transitted = List.map (fun eps ->
              { conf with
                Configuration.state = EpsilonTransition.next eps
              }
            ) epsilons in
            List.fold_right (fun (cfgs, rslts) (cfgs_rslt, rslts_rslt) ->
              ((cfgs @ cfgs_rslt), (rslts @ rslts_rslt))
            ) (List.map iter transitted) (cfgs_rslt, rslt_rslt)
          end
        ) var_processed ([], []) 
      in
          
      (finalized_cfgs @ epsilon_transitted, finalized_rslts @ epsilon_rslts)
*)
      (var_processed @ finalized_cfgs, finalized_rslts)
    in

    List.fold_right (fun (cfgs, rslts) (cfgs_rslt, rslts_rslt) ->
      ((cfgs @ cfgs_rslt), (rslts @ rslts_rslt))
    ) (List.map (fun cfg -> iter cfg) configurations) ([], [])
  in

  (* ---------------------------------------------------------------------- *)
(*  let rec movable_labels (cfg : Configuration.t) : SymbolSet.t =
    match cfg.Configuration.verif with
      | None ->
          let transitions = State.get_transitions cfg.Configuration.state in
          let rec loop lbls = function
            | [] -> lbls
            | x :: xs ->
                match Transition.symbol x with
                  | Symbol.SymStr str ->
                      loop (SymbolSet.add (Symbol.SymStr str) lbls) xs
                  | Symbol.SymAttr attr ->
                      loop (SymbolSet.add (Symbol.SymAttr attr) lbls) xs
                  | Symbol.SymLbl lbl ->
                      loop (SymbolSet.add (Symbol.SymLbl lbl) lbls) xs
                  | _ ->
                      loop lbls xs (* skip *)
          in
          loop SymbolSet.empty transitions
      | Some (_, x) ->
          match x with
            | XMLLeaf -> 
                SymbolSet.empty
            | XMLString (str, _, _) ->
                SymbolSet.singleton (Symbol.SymStr str)
            | XMLAttr (attr, _, _) ->
                SymbolSet.singleton (Symbol.SymAttr attr)
            | XMLNode (lbl, _, _) ->
                SymbolSet.singleton (Symbol.SymLbl lbl)
  in
*)

  (* 'convert' returns rest configurations to process and results *)
  let rec convert (configurations : Configuration.t list) (results : xml_tree list) : (Configuration.t list * xml_tree list) =
    (* prdbg_endline "================================================================================";
    prdbg_endline ("rest #cfg is " ^ (string_of_int (List.length configurations))); *)
    let (rest_configurations, stepped_results) = step configurations in
    (* prdbg_endline ("#stepped : " ^ (string_of_int (List.length rest_configurations))); *)
    if rest_configurations = [] then begin
      (rest_configurations, stepped_results @ results)
    end else begin
      convert rest_configurations (stepped_results @ results)
    end
  and step (configurations : Configuration.t list) : (Configuration.t list * xml_tree list) =
    (* prdbg_endline "[COMMAND] step"; *)
    let (configurations, results) = pre_transition configurations in
    let cfg_results = 
      List.flatten (List.map (fun cfg ->
        create_child_configuration cfg
      ) configurations)
    in
    (cfg_results, results)
  in

  (* ---------------------------------------------------------------------- *)

  let initial_configurations : Configuration.t list = create_initial_configurations input_binding in
  let (_, result) = convert initial_configurations [] in
  (* prdbg_endline ("#result is " ^ (string_of_int (List.length result))); *)
  match result with
    | [] -> failwith "No configuration left"
    | r :: _ -> r
;;
  
