(**
  @file    syntax.ml
  @brief   Syntax of XMLbitrans
  @author  shinyak
  @version $Rev: 268 $
  
  Shinya KAWANAKA (C) 2004 All Rights Reserved.
  $Id: syntax.ml 268 2005-04-24 11:38:59Z shinyak $
*)

open Util

type xml_tree =
  | XMLLeaf
  | XMLString of string * xml_tree * xml_tree            (* string *)
  | XMLAttr   of ident * xml_tree * xml_tree             (* *)
  | XMLNode   of ident * xml_tree * xml_tree             (* label, attributes, child, next *)


type xml_rev_tree =
  | XMLRevLeaf                                           (* leaf *)
  | XMLRevString of string * xml_rev_tree * xml_rev_tree (* string, leaf, prev *)
  | XMLRevAttr   of ident * xml_rev_tree * xml_rev_tree  
  | XMLRevNode   of ident * xml_rev_tree * xml_rev_tree  (* label, child, prev *)


let xml_tree_of_xml_rev_tree rev_tree =
  let rec iter after = function
    | XMLRevLeaf -> after
    | XMLRevString (str, child, prev) -> 
        iter (XMLString (str, iter XMLLeaf child, after)) prev;
    | XMLRevAttr (label, child, prev) ->
        iter (XMLAttr (label, iter XMLLeaf child, after)) prev
    | XMLRevNode (label, child, prev) ->
        iter (XMLNode (label, iter XMLLeaf child, after)) prev
  in
  iter XMLLeaf rev_tree
;;


let xml_rev_tree_of_xml_tree tree =
  let rec iter prev = function
    | XMLLeaf -> prev
    | XMLString (str, child, next) -> 
        iter (XMLRevString (str, iter XMLRevLeaf child, prev)) next
    | XMLAttr (label, child, next) ->
        iter (XMLRevAttr (label, iter XMLRevLeaf child, prev)) next
    | XMLNode (label, child, next) ->
        iter (XMLRevNode (label, iter XMLRevLeaf child, prev)) next
  in
  iter XMLRevLeaf tree
;;


type pattern = 
  | PEmptyElem
  | PAnyElem                        (* PAnyElem matches attribute *)
  | PAnyString
  | PString    of string
  | PAnyAttr
  | PAttr      of ident * pattern   (* pattern can be PStringAny or PString or PVar *)
  | PAttrLs    of ident             (* { x | x < ident } *)
  | PAttrGr    of ident             (* { x | ident < x } *)
  | PAttrLsGr  of ident * ident     (* { x | left < x < right } *)
  | PLabel     of ident * pattern
  | PStar      of pattern
  | PSeq       of pattern * pattern
  | POr        of pattern * pattern
  | PVar       of ident * pattern
  | PVarAny    of ident
;;


let rec string_of_pattern = function
  | PEmptyElem          -> ""
  | PAnyElem            -> "Any"
  | PAnyString          -> "String"
  | PString x           -> "STR:" ^ x
  | PAnyAttr            -> "AnyAttr"
  | PAttr (x, y)        -> "ATTR:(" ^ x ^ ", " ^ (string_of_pattern y) ^ " )"
  | PAttrLs ident       -> "ATTR:(x<" ^ ident ^ ")"
  | PAttrGr ident       -> "ATTR:(" ^ ident ^ "<x)"
  | PAttrLsGr (ls, gr)  -> "ATTR:(" ^ ls ^ "<x<" ^ gr ^ ")"
  | PLabel (ident, pat) -> "NOTE:(" ^ ident ^ "[" ^ (string_of_pattern pat) ^ "])"
  | PStar pat           -> "STAR:(" ^ (string_of_pattern pat) ^ ")"
  | PSeq (pat1, pat2)   -> "SEQ:(" ^ (string_of_pattern pat1) ^ " --> " ^ (string_of_pattern pat2) ^ ")"
  | POr (pat1, pat2)    -> "OR:(" ^ (string_of_pattern pat1) ^ " or " ^ (string_of_pattern pat2) ^ ")"
  | PVar (ident, pat)   -> "VAR:(" ^ ident ^ "[" ^ (string_of_pattern pat) ^ "])"
  | PVarAny ident       -> "VARANY:" ^ ident


type where_type = (ident * ident * ident) list;;  (* relname, var_left, var_right *)

type relation = {
  name  : ident;
  lhs   : pattern;
  rhs   : pattern;
  where : where_type;
};;

(*
let prdbg_relation relation =
  prdbg_string "relation : "; prdbg_endline relation.name;
  prdbg_string "lhs      : "; prdbg_endline (string_of_pattern relation.lhs);
  prdbg_string "rhs      : "; prdbg_endline (string_of_pattern relation.rhs);
  prdbg_string "where    : "; List.iter (fun (x, y, z) -> prdbg_string ("(" ^ x ^ ", " ^ y ^ ", " ^ z ^ ")")) relation.where;
  prdbg_newline ()
;;
*)

let rec varset_of_pattern = function
  | PEmptyElem
  | PAnyElem
  | PAnyString
  | PString _
  | PAnyAttr           -> StringSet.empty
  | PAttr (_, pat)     -> varset_of_pattern pat
  | PAttrLs _
  | PAttrGr _
  | PAttrLsGr _        -> StringSet.empty
  | PLabel (_, pat)    -> varset_of_pattern pat
  | PStar pat          -> varset_of_pattern pat
  | PSeq (pat1, pat2) 
  | POr (pat1, pat2)   -> StringSet.union (varset_of_pattern pat1) (varset_of_pattern pat2)
  | PVar (ident, pat)  -> StringSet.add ident (varset_of_pattern pat)
  | PVarAny ident      -> StringSet.singleton ident


let rec nullable is_left rel relations processed pattern = 
  let rec iter = function
    | PEmptyElem         -> true
    | PAnyElem           -> false
    | PAnyString         -> false
    | PString _          -> false
    | PAnyAttr           -> false
    | PAttr (_, pat)     -> false
    | PAttrLs _          -> false
    | PAttrGr _          -> false
    | PAttrLsGr _        -> false
    | PLabel (_, pat)    -> false
    | PStar pat          -> true
    | PSeq (pat1, pat2)  -> iter pat1 && iter pat2
    | POr (pat1, pat2)   -> iter pat1 || iter pat2
    | PVar (ident, pat)  -> iter pat
    | PVarAny ident      -> 
        if List.mem (rel.name, ident) processed then false 
        else begin
          let (am_name, _, _) =
            try
              List.find (fun (name, lhs, rhs) ->
                (is_left && lhs = ident) || ((not is_left) && rhs = ident)
              ) rel.where
            with
              | Not_found -> failwith ("Not found at check nullable:" ^ ident)
          in
          let relation = List.find (fun r -> r.name = am_name) relations in
          let pattern = if is_left then relation.lhs else relation.rhs in
          nullable is_left relation relations ((rel.name, ident) :: processed) pattern 
        end
  in
  iter pattern
;;

let check_nullable relations =
  let rec check is_left rel = function
    | PEmptyElem         -> true
    | PAnyElem           -> true
    | PAnyString         -> true
    | PString _          -> true
    | PAnyAttr           -> true
    | PAttr (_, pat)     -> check is_left rel pat
    | PAttrLs _          -> true
    | PAttrGr _          -> true
    | PAttrLsGr _        -> true
    | PLabel (_, pat)    -> check is_left rel pat
    | PStar pat          -> if nullable is_left rel relations [] pat then false else true
    | PSeq (pat1, pat2)  -> check is_left rel pat1 && check is_left rel pat2
    | POr (pat1, pat2)   -> check is_left rel pat1 && check is_left rel pat2
    | PVar (ident, pat)  -> true
    | PVarAny ident      -> true
  in
  List.iter (fun rel ->
    let lhs = rel.lhs
    and rhs = rel.rhs in
    if not (check true  rel lhs) then failwith "check_nullable";
    if not (check false rel rhs) then failwith "check_nullable";
  ) relations
;;
(* type where_type = (ident * ident * ident) list;;  (* relname, var_left, var_right *) *)



